/*
 *	@(#)CompressionStreamElement.java 1.4 99/08/25 17:21:41
 *
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 */

package com.sun.j3d.utils.compression ;

/**
 * Instances of this class are used as elements in a CompressionStream.
 * @see CompressionStream
 */
abstract class CompressionStreamElement {
    /**
     * Bit length of quantized geometric components.
     */
    int length ;

    /**
     * Number of trailing zeros in quantized geometric components.
     */
    int shift ;

    /**
     * If false, geometric component values are represented as differences
     * from those of the preceding element in the stream.
     */
    boolean absolute ;

    /**
     * Array with elements that can be used as masks to apply a quantization
     * to the number of bits indicated by the referencing index [0..16].
     */
    static final int quantizationMask[] = {
	0xFFFF0000, 0xFFFF8000, 0xFFFFC000, 0xFFFFE000,
	0xFFFFF000, 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00,
	0xFFFFFF00, 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0,
	0xFFFFFFF0, 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE,
	0xFFFFFFFF
    } ;
    
    /**
     * Array with elements that can be used as masks to retain the number of
     * trailing bits of data indicated by the referencing index [0..64].  Used
     * to clear the leading sign bits of fixed-point 2's complement numbers
     * and in building the compressed output stream.
     */
    static final long lengthMask[] = {
	0x0000000000000000L, 0x0000000000000001L,
	0x0000000000000003L, 0x0000000000000007L,
	0x000000000000000FL, 0x000000000000001FL,
	0x000000000000003FL, 0x000000000000007FL,
	0x00000000000000FFL, 0x00000000000001FFL,
	0x00000000000003FFL, 0x00000000000007FFL,
	0x0000000000000FFFL, 0x0000000000001FFFL,
	0x0000000000003FFFL, 0x0000000000007FFFL,
	0x000000000000FFFFL, 0x000000000001FFFFL,
	0x000000000003FFFFL, 0x000000000007FFFFL,
	0x00000000000FFFFFL, 0x00000000001FFFFFL,
	0x00000000003FFFFFL, 0x00000000007FFFFFL,
	0x0000000000FFFFFFL, 0x0000000001FFFFFFL,
	0x0000000003FFFFFFL, 0x0000000007FFFFFFL,
	0x000000000FFFFFFFL, 0x000000001FFFFFFFL,
	0x000000003FFFFFFFL, 0x000000007FFFFFFFL,
	0x00000000FFFFFFFFL, 0x00000001FFFFFFFFL,
	0x00000003FFFFFFFFL, 0x00000007FFFFFFFFL,
	0x0000000FFFFFFFFFL, 0x0000001FFFFFFFFFL,
	0x0000003FFFFFFFFFL, 0x0000007FFFFFFFFFL,
	0x000000FFFFFFFFFFL, 0x000001FFFFFFFFFFL,
	0x000003FFFFFFFFFFL, 0x000007FFFFFFFFFFL,
	0x00000FFFFFFFFFFFL, 0x00001FFFFFFFFFFFL,
	0x00003FFFFFFFFFFFL, 0x00007FFFFFFFFFFFL,
	0x0000FFFFFFFFFFFFL, 0x0001FFFFFFFFFFFFL,
	0x0003FFFFFFFFFFFFL, 0x0007FFFFFFFFFFFFL,
	0x000FFFFFFFFFFFFFL, 0x001FFFFFFFFFFFFFL,
	0x003FFFFFFFFFFFFFL, 0x007FFFFFFFFFFFFFL,
	0x00FFFFFFFFFFFFFFL, 0x01FFFFFFFFFFFFFFL,
	0x03FFFFFFFFFFFFFFL, 0x07FFFFFFFFFFFFFFL,
	0x0FFFFFFFFFFFFFFFL, 0x1FFFFFFFFFFFFFFFL,
	0x3FFFFFFFFFFFFFFFL, 0x7FFFFFFFFFFFFFFFL,
	0xFFFFFFFFFFFFFFFFL
    } ;


    /**
     * Computes the quantized representation of this stream element.
     *
     * @param stream CompressionStream associated with this element
     * @param table HuffmanTable for collecting data about the quantized
     * representation of this element
     */
    abstract void quantize(CompressionStream stream, HuffmanTable table) ;

    /**
     * Outputs the compressed bits representing this stream element.
     * Some instances of CompressionStreamElement don't require an
     * implementation and will inherit the stub provided here.
     *
     * @param table HuffmanTable mapping quantized representations to
     * compressed encodings
     * @param output CommandStream for collecting compressed output
     */
    void outputCommand(HuffmanTable table, CommandStream output) {
    }

    /**
     * Finds the minimum bits needed to represent the given signed 2's
     * complement number.  For positive numbers, this include the first 1
     * starting from the left, plus a 0 sign bit; for negative numbers, this
     * includes the first 0 starting from the left, plus a 1 sign bit.  The
     * numbers 0 and -1 are special cases returning lengths of 1.<p>
     * 
     * The decompressor uses the data length to determine how many bits of
     * sign extension to add to the data coming in from the compressed stream
     * in order to create a 16-bit signed 2's complement number.  E.g., a data
     * length of 12 indicates that 16-12=4 bits of sign are to be extended.<p>
     *
     * @param number a signed 2's complement number representable in 16 bits
     * or less
     * @return minimum number of bits to represent the number
     */
    static final int getLength(int number) {
	if (number == 0 || number == -1)
	    return 1 ;

	int i ;
	int signBit = (number & 0x8000) >> 15 ;

	for (i = 14 ; i > 0 ; i--)
	    if (((number >> i) & 0x1) != signBit)
		break ;

	return i + 2 ;
    }

    /**
     * Finds the rightmost 1 bit in the given signed 2's complement number.
     * 0 is treated as a special case returning 0.<p>
     * 
     * This value is used by the decompressor to indicate the number of
     * trailing zeros to be added to the end of the data coming in from the
     * compressed stream, accomplished by left shifting the data by the
     * indicated amount.<p>
     *
     * @param number a signed 2's complement number representable in 16 bits
     * or less
     * @return number of trailing zeros
     */
    static final int getShift(int number) {
	if (number == 0)
	    return 0 ;

	int i ;
	for (i = 0 ; i < 15 ; i++)
	    if ((number & (1 << i)) != 0)
		break ;

	return i ;
    }

    /**
     * Finds the maximum of two integers.
     */
    static final int getMaximum(int x, int y) {
	if (x > y)
	    return x ;
	else
	    return y ;
    }

    /**
     * Finds the minimum of two integers.
     */
    static final int getMinimum(int x, int y) {
	return -getMaximum(-x, -y) ;
    }

    /**
     * Finds the maximum of three integers.
     */
    static final int getMaximum(int x, int y, int z) {
	return getMaximum(getMaximum(x, y), z) ;
    }

    /**
     * Finds the minimum of three integers.
     */
    static final int getMinimum(int x, int y, int z) {
	return -getMaximum(-x, -y, -z) ;
    }

    /**
     * Finds the maximum of four integers.
     */
    static final int getMaximum(int x, int y, int z, int w) {
	return getMaximum(getMaximum(x, y), getMaximum(z, w)) ;
    }

    /**
     * Finds the minimum of four integers.
     */
    static final int getMinimum(int x, int y, int z, int w) {
	return -getMaximum(-x, -y, -z, -w) ;
    }
}

